@@ -20,7 +20,9 @@ are what gets used.
 
 from array import array
 from itertools import count
-import re, time
+import re
+import sys
+import time
 try:
     import cPickle as pickle
 except ImportError:
@@ -31,6 +33,7 @@ from classregistry import findClass
 # Sadly the name "constraints" conflicts with many of the function
 # arguments in this module, so we rename it:
 import constraints as constrs
+import converters
 import sqlbuilder
 from styles import capword
 
@@ -71,6 +74,24 @@ if mxdatetime_available:
     __all__.append("MXDATETIME_IMPLEMENTATION")
 
 
+def use_microseconds(use=True):
+    if use:
+        SODateTimeCol.datetimeFormat = '%Y-%m-%d %H:%M:%S.%f'
+        SOTimeCol.timeFormat = '%H:%M:%S.%f'
+        dt_types = [(datetime.datetime, converters.DateTimeConverterMS),
+                    (datetime.time, converters.TimeConverterMS)]
+    else:
+        SODateTimeCol.datetimeFormat = '%Y-%m-%d %H:%M:%S'
+        SOTimeCol.timeFormat = '%H:%M:%S'
+        dt_types = [(datetime.datetime, converters.DateTimeConverter),
+                    (datetime.time, converters.TimeConverter)]
+    for dt_type, converter in dt_types:
+        converters.registerConverter(dt_type, converter)
+
+
+__all__.append("use_microseconds")
+
+
 creationOrder = count()
 
 ########################################
@@ -318,7 +339,8 @@ class SOCol(object):
     def _maxdbType(self):
         return self._sqlType()
 
-    def mysqlCreateSQL(self):
+    def mysqlCreateSQL(self, connection=None):
+        self.connection = connection
         return ' '.join([self.dbName, self._mysqlType()] + self._extraSQL())
 
     def postgresCreateSQL(self):
@@ -579,9 +601,9 @@ class UnicodeStringValidator(SOValidator):
         if isinstance(value, (unicode, sqlbuilder.SQLExpression)):
             return value
         if isinstance(value, str):
-            return unicode(value, self.getDbEncoding(state))
+            return value.decode(self.getDbEncoding(state))
         if isinstance(value, array): # MySQL
-            return unicode(value.tostring(), self.getDbEncoding(state))
+            return value.tostring().decode(self.getDbEncoding(state))
         if hasattr(value, '__unicode__'):
             return unicode(value)
         raise validators.Invalid("expected a str or a unicode in the UnicodeCol '%s', got %s %r instead" % \
@@ -796,21 +818,51 @@ class SOKeyCol(SOCol):
         self.refColumn = kw.pop('refColumn', None)
         super(SOKeyCol, self).__init__(**kw)
 
+    def _idType(self):
+        return self.soClass.sqlmeta.idType
+
     def _sqlType(self):
-        return self.key_type[self.soClass.sqlmeta.idType]
+        return self.key_type[self._idType()]
 
     def _sybaseType(self):
         key_type = {int: "NUMERIC(18,0) NULL", str: "TEXT"}
-        return key_type[self.soClass.sqlmeta.idType]
+        return key_type[self._idType()]
 
     def _mssqlType(self):
         key_type = {int: "INT NULL", str: "TEXT"}
-        return key_type[self.soClass.sqlmeta.idType]
+        return key_type[self._idType()]
 
 class KeyCol(Col):
 
     baseClass = SOKeyCol
 
+class ForeignKeyValidator(SOValidator):
+
+    def __init__(self, *args, **kw):
+        super(ForeignKeyValidator, self).__init__(*args, **kw)
+        self.fkIDType = None
+
+    def from_python(self, value, state):
+        if value is None:
+            return None
+        # Avoid importing the main module
+        # to get the SQLObject class for isinstance
+        if hasattr(value, 'sqlmeta'):
+            return value
+        if self.fkIDType is None:
+            otherTable = findClass(self.soCol.foreignKey,
+                                   self.soCol.soClass.sqlmeta.registry)
+            self.fkIDType = otherTable.sqlmeta.idType
+        try:
+            value = self.fkIDType(value)
+            return value
+        except (ValueError, TypeError):
+            pass
+        raise validators.Invalid("expected a %r for the ForeignKey '%s', "
+                                 "got %s %r instead" %
+                                 (self.fkIDType, self.name,
+                                  type(value), value), value, state)
+
 class SOForeignKey(SOKeyCol):
 
     def __init__(self, **kw):
@@ -823,6 +875,14 @@ class SOForeignKey(SOKeyCol):
             kw['name'] = style.instanceAttrToIDAttr(style.pythonClassToAttr(foreignKey))
         super(SOForeignKey, self).__init__(**kw)
 
+    def createValidators(self):
+        return [ForeignKeyValidator(name=self.name)] + \
+            super(SOForeignKey, self).createValidators()
+
+    def _idType(self):
+        other = findClass(self.foreignKey, self.soClass.sqlmeta.registry)
+        return other.sqlmeta.idType
+
     def sqliteCreateSQL(self):
         sql = SOKeyCol.sqliteCreateSQL(self)
         other = findClass(self.foreignKey, self.soClass.sqlmeta.registry)
@@ -904,8 +964,8 @@ class SOForeignKey(SOKeyCol):
                        'sTLocalName': sTLocalName})
         return constraint
 
-    def mysqlCreateSQL(self):
-        return SOKeyCol.mysqlCreateSQL(self)
+    def mysqlCreateSQL(self, connection=None):
+        return SOKeyCol.mysqlCreateSQL(self, connection)
 
     def sybaseCreateSQL(self):
         sql = SOKeyCol.sybaseCreateSQL(self)
@@ -1098,11 +1158,23 @@ class DateTimeValidator(validators.DateValidator):
                 else:
                     return datetime.time(value.hour, value.minute, int(value.second))
         try:
-            stime = time.strptime(value, self.format)
+            if self.format.find(".%f") >= 0:
+                if '.' in value:
+                    _value = value.split('.')
+                    microseconds = _value[-1]
+                    _l = len(microseconds)
+                    if _l < 6:
+                        _value[-1] = microseconds + '0'*(6 - _l)
+                    elif _l > 6:
+                        _value[-1] = microseconds[:6]
+                    if _l != 6:
+                        value = '.'.join(_value)
+                else:
+                    value += '.0'
+            return datetime.datetime.strptime(value, self.format)
         except:
             raise validators.Invalid("expected a date/time string of the '%s' format in the DateTimeCol '%s', got %s %r instead" % \
                 (self.format, self.name, type(value), value), value, state)
-        return datetime.datetime(*stime[:6])
 
     def from_python(self, value, state):
         if value is None:
@@ -1129,11 +1201,25 @@ if mxdatetime_available:
             elif isinstance(value, datetime.time):
                 return DateTime.Time(value.hour, value.minute, value.second)
             try:
-                stime = time.strptime(value, self.format)
+                if self.format.find(".%f") >= 0:
+                    if '.' in value:
+                        _value = value.split('.')
+                        microseconds = _value[-1]
+                        _l = len(microseconds)
+                        if _l < 6:
+                            _value[-1] = microseconds + '0'*(6 - _l)
+                        elif _l > 6:
+                            _value[-1] = microseconds[:6]
+                        if _l != 6:
+                            value = '.'.join(_value)
+                    else:
+                        value += '.0'
+                value = datetime.datetime.strptime(value, self.format)
+                return DateTime.DateTime(value.year, value.month, value.day,
+                    value.hour, value.minute, value.second)
             except:
                 raise validators.Invalid("expected a date/time string of the '%s' format in the DateTimeCol '%s', got %s %r instead" % \
                     (self.format, self.name, type(value), value), value, state)
-            return DateTime.mktime(stime)
 
         def from_python(self, value, state):
             if value is None:
@@ -1146,7 +1232,7 @@ if mxdatetime_available:
                 (self.name, type(value), value), value, state)
 
 class SODateTimeCol(SOCol):
-    datetimeFormat = '%Y-%m-%d %H:%M:%S'
+    datetimeFormat = '%Y-%m-%d %H:%M:%S.%f'
 
     def __init__(self, **kw):
         datetimeFormat = kw.pop('datetimeFormat', None)
@@ -1165,7 +1251,10 @@ class SODateTimeCol(SOCol):
         return _validators
 
     def _mysqlType(self):
-        return 'DATETIME'
+        if self.connection and self.connection.can_use_microseconds():
+            return 'DATETIME(6)'
+        else:
+            return 'DATETIME'
 
     def _postgresType(self):
         return 'TIMESTAMP'
@@ -1174,7 +1263,10 @@ class SODateTimeCol(SOCol):
         return 'DATETIME'
 
     def _mssqlType(self):
-        return 'DATETIME'
+        if self.connection and self.connection.can_use_microseconds():
+            return 'DATETIME2(6)'
+        else:
+            return 'DATETIME'
 
     def _sqliteType(self):
         return 'TIMESTAMP'
@@ -1277,7 +1369,7 @@ class TimeValidator(DateTimeValidator):
     from_python = to_python
 
 class SOTimeCol(SOCol):
-    timeFormat = '%H:%M:%S'
+    timeFormat = '%H:%M:%S.%f'
 
     def __init__(self, **kw):
         timeFormat = kw.pop('timeFormat', None)
@@ -1296,7 +1388,10 @@ class SOTimeCol(SOCol):
         return _validators
 
     def _mysqlType(self):
-        return 'TIME'
+        if self.connection and self.connection.can_use_microseconds():
+            return 'TIME(6)'
+        else:
+            return 'TIME'
 
     def _postgresType(self):
         return 'TIME'
@@ -1304,6 +1399,12 @@ class SOTimeCol(SOCol):
     def _sybaseType(self):
         return 'TIME'
 
+    def _mssqlType(self):
+        if self.connection and self.connection.can_use_microseconds():
+            return 'TIME(6)'
+        else:
+            return 'TIME'
+
     def _sqliteType(self):
         return 'TIME'
 
@@ -1328,7 +1429,10 @@ class SOTimestampCol(SODateTimeCol):
         SOCol.__init__(self, **kw)
 
     def _mysqlType(self):
-        return 'TIMESTAMP'
+        if self.connection and self.connection.can_use_microseconds():
+            return 'TIMESTAMP(6)'
+        else:
+            return 'TIMESTAMP'
 
 class TimestampCol(Col):
     baseClass = SOTimestampCol
